# -*- coding: utf-8 -*-
"""
Data modeling for command history to be modular
"""
from __future__ import print_function
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
import os
import sys

import requests


class HistoryBase(object):
    """
    Base class of what all backend command history
    searches should use.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def search_reverse(self, term):
        """
        Return a list of commmands that is in reverse
        time order. i.e newest first.
        """
        pass


class LocalHistory(HistoryBase):
    """
    Use local .bash_history for doing searches
    """
    def __init__(self):
        """
        Load up the bash history uniqueified into an
        OrderedDict for forward/backward searching and then
        dumped to a list.
        """
        history_dict = OrderedDict()
        with open(os.path.expanduser('~/.bash_history')) as history_file:
            for line in history_file:
                history_dict[line.strip()] = None
        self.data = history_dict.keys()

    def search_reverse(self, term):
        """
        Return reversed filtered list by term
        """
        results = [
            x for x in self.data
            if term in x
        ]
        results.reverse()
        return results


class WebHistory(HistoryBase):
    """
    Use RESTful API to do searches against archelond.
    """
    SEARCH_URL = '/api/v1/history'

    def __init__(self, url, token):
        """
        Setup requests session with API key and set base
        URL.
        """
        self.url = '{url}{endpoint}'.format(
            url=url.rstrip('/'),
            endpoint=self.SEARCH_URL
        )
        self.session = requests.Session()
        self.session.headers = {'Authorization': 'token {}'.format(token)}

    def search_reverse(self, term):
        """
        Make request to API with sort order specified
        and return the results as a list.
        """
        try:
            response = self.session.get(
                self.url,
                params={'q': term, 'o': 'r'}
            )
        except requests.exceptions.ConnectionError:
            print('Failed to connect to server, check settings')
            sys.exit(1)

        if response.status_code != 200:
            return ['Error in API Call {}'.format(response.text)]
        return [x['command'] for x in response.json()['commands']]

    def add(self, command):
        """
        Post a command to the remote server using the API
        """
        try:
            response = self.session.post(
                self.url,
                json={'command': command}
            )
        except requests.exceptions.ConnectionError:
            print('Failed to connect to server, check settings')
            sys.exit(1)

        if response.status_code != 201:
            return False, (response.json(), response.status_code)
        else:
            return True, None

    def bulk_add(self, commands):
        """
        Post a list of commands
        """
        response = self.session.post(
            self.url,
            json={'commands': commands}
        )
        if response.status_code != 200:
            return False, (response.json(), response.status_code)
        else:
            return True, (response.json(), response.status_code)



# -*- coding: utf-8 -*-
"""
Command line entry points for the archelon client.
"""
from __future__ import print_function
from difflib import Differ
import os
import shutil
import sys

import requests

from archelonc.search import Search
from archelonc.data import WebHistory

HISTORY_FILE = os.path.expanduser('~/.archelon_history')


def _get_web_setup():
    """
    either get a WebHistory object or None if we aren't configured for
    one.
    """
    # Check if we are pointed at an archelond server
    url = os.environ.get('ARCHELON_URL')
    token = os.environ.get('ARCHELON_TOKEN')
    if not (url and token):
        return None
    return WebHistory(url, token)


def search_form():
    """
    Entry point to search history
    """
    Search().run()


def update():
    """
    Capture diff between stored history and new history and upload
    the delta.
    """
    web_history = _get_web_setup()
    # If we aren't setup for Web usage, just bomb out.
    if not web_history:
        sys.exit(0)

    current_hist_file = os.path.expanduser(
        os.environ.get('HISTFILE', '~/.bash_history')
    )
    # Create our diff file if it doesn't exist
    if not os.path.exists(HISTORY_FILE):
        open(HISTORY_FILE, 'a').close()

    # Compare the current history to our previously stored one,
    # upload any additions and copy the file over.
    commands = {}
    with open(HISTORY_FILE) as cached, open(current_hist_file) as current:
        differ = Differ()
        results = differ.compare(cached.readlines(), current.readlines())

        # use diff lib "codes" to see if we need to upload differences
        for diff in results:
            if diff[:2] == '+ ' or diff[:2] == '? ':
                if diff[:2]:
                    commands[diff[2:]] = None

    # Warn if we are doing a large upload
    num_commands = len(commands.keys())
    if num_commands > 50:
        print('Beginning upload of {} history items. '
              'This may take a while...'.format(num_commands))

    try:
        success, response = web_history.bulk_add(commands.keys())
    except requests.exceptions.ConnectionError, ex:
        print('Connection Error occured: %s', str(ex))
        sys.exit(1)
    if not success:
        print('Failed to add commands, got:\n {}'.format(
            response
        ))
        sys.exit(1)
    shutil.copy(current_hist_file, HISTORY_FILE)


def import_history():
    """
    Import current shell's history into server
    """
    web_history = _get_web_setup()
    if not web_history:
        print(
            'You must specify an archelon server in the environment with '
            '`ARCHELON_URL` and `ARCHELON_TOKEN` environment variables'
        )
        sys.exit(1)

    # Read history and post to server.  if arg[1]
    # is given, use that file as the history file.
    hist_file = os.environ.get('HISTFILE', '~/.bash_history')
    if len(sys.argv) == 2:
        hist_file = sys.argv[1]

    hist_file_path = os.path.expanduser(hist_file)
    with open(hist_file_path) as history_file:
        commands = {}
        for line in history_file:
            command = line.strip()
            if not command:
                continue
            commands[command] = None
    try:
        success, response = web_history.bulk_add(commands.keys())
    except requests.exceptions.ConnectionError, ex:
        print('Connection Error occured: %s', str(ex))
        sys.exit(1)

    if not success:
        print('Failed to add commands, got:\n {}'.format(
            response
        ))
    # Make copy of imported history so we only have to track
    # changes from here on out when archelon is invoked
    shutil.copy(hist_file_path, HISTORY_FILE)



# -*- coding: utf-8 -*-
"""
npyscreen based application for searching shell history
"""
import curses
import os
import sys

import npyscreen

from archelonc.data import LocalHistory, WebHistory


class SearchResult(npyscreen.Textfield):
    """
    Search result item
    """
    pass


class CommandBox(npyscreen.TitleText):
    """
    Command Box widget
    """

    def __init__(self, screen, **kwargs):
        """
        Just add some state to the default
        widget to track if we have been manually
        edited
        """
        self.been_edited = False
        super(CommandBox, self).__init__(screen, **kwargs)

    def when_value_edited(self):
        """
        Mark myself as having been edited
        """
        self.been_edited = True


class SearchResults(npyscreen.MultiLineAction):
    """
    MultiLine widget for displaying search
    results.
    """
    _contained_widgets = SearchResult

    def actionHighlighted(self, act_on_this, key_press):
        cmd_box = self.parent.command_box
        cmd_box.value = act_on_this
        cmd_box.been_edited = True
        cmd_box.update()
        self.parent.editw = 2
        self.parent.edit()


class SearchBox(npyscreen.TitleText):
    """
    Search box command, updates trigger
    deeper searching.
    """
    def search(self):
        """
        Do the search and return the results
        """
        return self.parent.parentApp.data.search_reverse(self.value)

    def when_value_edited(self):
        """
        Do the search and filter the search result
        list based on what is returned
        """
        if len(self.value) == 0:
            return
        results_list = self.parent.results_list
        cmd_box = self.parent.command_box

        search_results = self.search()
        results_list.values = search_results
        results_list.reset_display_cache()
        results_list.reset_cursor()
        results_list.update()

        # If you haven't edited the command box, go ahead
        # and put the top value in there to save time.
        if not cmd_box.been_edited:
            if len(search_results) > 0:
                cmd_box.value = search_results[0]
                cmd_box.update()


class SearchReverse(npyscreen.ActionFormV2):
    """
    Reverse command history search.
    """

    def afterEditing(self):
        """
        This is the form to display, so set next to None
        """
        self.parentApp.setNextForm(None)

    def create(self):
        """
        Build the form for searching
        """
        self.results_list = self.add(
            SearchResults,
            name='Results',
            scroll_exit=True,
            max_height=-6,
            values=[]
        )
        self.search_box = self.add(
            SearchBox,
            name='Search',
            rely=-4,
            begin_entry_at=10
        )
        self.command_box = self.add(
            CommandBox,
            name='Command',
            rely=-3,
            begin_entry_at=10
        )

        self.add_handlers({
            '!o': self.on_ok,
            '!c': self.on_cancel,
            curses.ascii.ESC: self.on_cancel
        })

    def beforeEditing(self):
        """
        Set the edit index to the search box
        and tell it to preserve the value
        """
        self.editw = 1
        self.preserve_selected_widget = True

    def on_ok(self, *args):
        """
        We just drop the command into a
        known file for the wrapper to pick
        up
        """
        with open(os.path.expanduser('~/.archelon_cmd'), 'w') as cmd_file:
            cmd_file.write(self.command_box.value)
        sys.exit(0)

    def on_cancel(self, *args):
        """
        Drop out with a non 1 exit code so the
        wrapper doesn't execute anything
        """
        sys.exit(1)


class Search(npyscreen.NPSAppManaged):
    """
    Search application.  Determines which form to show.
    """
    def onStart(self):
        """
        Startup routine for the search application
        """
        url = os.environ.get('ARCHELON_URL')
        token = os.environ.get('ARCHELON_TOKEN')

        #  Determine the data model to use.
        if url and token:
            self.data = WebHistory(url, token)
        else:
            self.data = LocalHistory()

        self.addForm('MAIN', SearchReverse, name='Archelon: Reverse Search')



#!/usr/bin/env python
"""
Package installer for archelon client
"""

from setuptools import setup, find_packages


VERSION = __import__('archelonc').VERSION

with open('README.rst') as readme:
    README = readme.read()

setup(
    name='archelonc',
    version=VERSION,
    packages=find_packages(),
    package_data={},
    license='AGPLv3',
    author='Carson Gee',
    author_email='x@carsongee.com',
    url="http://github.com/carsongee/arcelon",
    description=("Client connected to archelonc for Web shell history"),
    long_description=README,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Information Technology',
        ('License :: OSI Approved :: '
         'GNU Affero General Public License v3'),
        'Operating System :: POSIX :: Linux',
    ],
    install_requires=[
        'npyscreen',
        'requests',
        ],
    scripts=['scripts/archelon'],
    entry_points={'console_scripts': [
        'archelonf = archelonc.command:search_form',
        'archelon_update = archelonc.command:update',
        'archelon_import = archelonc.command:import_history',
    ]},
    zip_safe=True,
)
